import argparse
import os
import rpm
import sys
import getpass

from pyrpkg.cli import cliClient
from pyrpkg import utils

from .exceptions import NotUnpackedException, RpmSpecParseException

class rpkgClient(cliClient):
    def __init__(self, config, name=None):
        self.DEFAULT_CLI_NAME = 'rpkg'

        self.auto_pack = False
        self.preprocess = True
        self.base_output_path = '/tmp/rpkg'
        self.ns_module_name = None

        if config:
            self.auto_pack = config['rpkg'].getboolean('auto_pack', False)
            self.preprocess = config['rpkg'].getboolean('preprocess_spec', True)
            self.base_output_path = config['rpkg'].get('base_output_path', '/tmp/rpkg')
            self.ns_module_name = config['rpkg'].get('module_name', None)

        if not os.path.isdir(self.base_output_path):
            prev_umask = os.umask(0)
            os.makedirs(self.base_output_path, mode=0o1777)
            os.umask(prev_umask)

        super(rpkgClient, self).__init__(config, name)

    def setup_argparser(self):
        """Setup the argument parser and register some basic commands."""
        self.spec_pick_parser = argparse.ArgumentParser(add_help=False)
        self.spec_pick_parser.add_argument('--outdir', action='store', default=None,
                                           help='Where to put any generated rpm content. '
                                           'Directory must exist.')
        self.spec_pick_parser.add_argument('--spec', action='store', default=None,
                                           help='An input spec template. If not given, '
                                           'the spec template is auto-located in path as a '
                                           'file ending with ".spec.rpkg" or ".spec" suffix '
                                           'in this order.')

        self.parser = argparse.ArgumentParser(
            prog=self.name, epilog='For detailed help pass --help to a target.')

        self.parser.add_argument('--config', '-C', default=None,
                                 help='Specify a config file to use.')
        self.parser.add_argument('--module-name',
                                 help='Override the module name. Otherwise '
                                 'it is discovered from: Git push URL '
                                 'or Git URL.', default=self.ns_module_name)
        self.parser.add_argument('--user', default=None,
                                 help='Override the discovered user name.')
        self.parser.add_argument('--path', action='store', default=None, type=utils.u,
                                 help='Define the directory to work in (defaults to cwd).')
        # Verbosity
        self.parser.add_argument('--verbose', '-v', dest='v',
                                 action='store_true',
                                 help='Run with verbose debug output.')
        self.parser.add_argument('--debug', '-d', dest='debug',
                                 action='store_true',
                                 help='Run with debug output.')
        self.parser.add_argument('-q', action='store_true',
                                 help='Run quietly only displaying errors.')

    def setup_subparsers(self):
        """Setup basic subparsers that all clients should use"""

        # Setup some basic shared subparsers

        # help command
        self.register_help()

        # Add a common parsers
        self.register_rpm_common()

        # Other targets
        self.register_make_source()
        self.register_clean()
        self.register_clog()
        self.register_clone()
        self.register_commit()
        self.register_compile()
        self.register_copr_build()
        self.register_diff()
        self.register_gimmespec()
        self.register_import_srpm()
        self.register_install()
        self.register_is_packed()
        self.register_lint()
        self.register_local()
        self.register_new()
        self.register_new_sources()
        self.register_prep()
        self.register_pull()
        self.register_push()
        self.register_sources()
        self.register_spec()
        self.register_srpm()
        self.register_switch_branch()
        self.register_tag()
        self.register_unused_patches()
        self.register_upload()
        self.register_verify_files()
        self.register_verrel()

        self.extend_subparsers()

    def extend_subparsers(self):
        spec_pick_cmds = ['make-source', 'compile', 'copr-build',
                          'build', 'install', 'is-packed',
                          'local', 'prep', 'spec', 'srpm', 'tag',
                          'unused-patches', 'verrel', 'clog']

        for spec_pick_cmd in spec_pick_cmds:
            self.subparsers.add_parser(name=spec_pick_cmd, add_help=False,
                description=self.subparsers.choices[spec_pick_cmd].description,
                parents=[self.subparsers.choices[spec_pick_cmd], self.spec_pick_parser])

    def load_cmd(self):
        """This sets up the cmd object"""

        # load git items from the config file
        items = {}
        if self.config.has_section('git'):
            items = dict(self.config.items('git', raw=True))

        # Read comma separated list of kerberos realms
        realms = [realm
                  for realm in items.get("kerberos_realms", '').split(',')
                  if realm]

        # Create the cmd object
        self._cmd = self.site.Commands(self.args.path,
                                       items.get('lookaside', ''),
                                       items.get('lookasidehash', 'sha512'),
                                       items.get('lookaside_cgi', ''),
                                       items.get('gitbaseurl', ''),
                                       items.get('anongiturl', ''),
                                       branchre='.*',
                                       kojiconfig='',
                                       build_client=None,
                                       user=self.args.user,
                                       quiet=self.args.q,
                                       realms=realms,
                                       outdir=getattr(self.args, 'outdir', None),
                                       preprocess=self.preprocess,
                                       base_output_path=self.base_output_path)

        self._cmd.ns_module_name = self.args.module_name
        self._cmd.debug = self.args.debug
        self._cmd.verbose = self.args.v
        self._cmd.clone_config = items.get('clone_config')
        self._cmd.set_specloc(getattr(self.args, 'spec', None))

    def clog(self):
        self.cmd.dump_spec_only = True
        self.cmd.clog(raw=self.args.raw)

    def lint(self):
        self.cmd.use_existing_outdir = True
        self.cmd.lint(self.args.info, self.args.rpmlintconf)

    def install(self):
        self.sources()
        self.cmd.install(target=self.args.target,
                         short=self.args.short_circuit,
                         nocheck=self.args.nocheck,
                         bcond_with=getattr(self.args, 'with'),
                         bcond_without=self.args.without)

    def compile(self):
        self.sources()
        self.cmd.compile(target=self.args.target,
                         short=self.args.short_circuit,
                         nocheck=self.args.nocheck,
                         bcond_with=getattr(self.args, 'with'),
                         bcond_without=self.args.without)

    def prep(self):
        self.sources()
        self.cmd.prep(target=self.args.target,
                      bcond_with=getattr(self.args, 'with'),
                      bcond_without=self.args.without)

    def local(self):
        self.sources()
        self.cmd.local(target=self.args.target,
                       bcond_with=getattr(self.args, 'with'),
                       bcond_without=self.args.without)

    def push(self):
        self.cmd.push(getattr(self.args, 'force', False))

    def verify_files(self):
        self.cmd.use_existing_outdir = True
        self.cmd.verify_files()

    def verrel(self):
        self.cmd.dump_spec_only = True
        print('%s-%s-%s' % (self.cmd.module_name, self.cmd.ver,
                            self.cmd.rel))

    def unused_patches(self):
        self.cmd.dump_spec_only = True
        unused = self.cmd.unused_patches()
        print('\n'.join(unused))

    def make_source(self):
        sys.stderr.write('This function is deprecated and will be removed in a future release.\n')
        self.cmd.sources()
        self.cmd.make_source()

    def srpm(self):
        self.cmd.sources()
        if self.auto_pack:
            try:
                self.cmd.make_source()
            except NotUnpackedException:
                pass
            else:
                sys.stderr.write('auto-packing: This function is deprecated and will be removed in a future release.\n')
        return self.cmd.srpm(target=getattr(self.args, 'target', None),
                             bcond_with=getattr(self.args, 'with', []),
                             bcond_without=getattr(self.args, 'without', []))

    def tag(self):
        self.cmd.dump_spec_only = True

        if not self.args.tag:
            self.cmd.version_bump = True

        self.cmd._rpmdefines = self.cmd.rpmdefines + ["--define 'dist %nil'"]

        if self.args.list:
            self.cmd.list_tag(self.args.tag)
        elif self.args.delete:
            if not self.args.tag:
                sys.stderr.write('Please, specify tag to be deleted.\n')
                return
            else:
                self.cmd.delete_tag(self.args.tag)
        else:
            self.cmd.add_tag(self.args.tag, self.args.skip_edit)

    def gimmespec(self):
        print(self.cmd.repo_spec)

    def spec(self):
        if not self.args.sources:
            self.cmd.dump_spec_only = True
        self.cmd.load_spec()

    def commit(self):
        if self.args.with_changelog and not self.args.message:
            raise rpkgError('--with-changelog must be used with -m together.')

        if self.args.message and self.args.with_changelog:
            # Compose commit message with a summary and content into a file.
            self.cmd.clog(True)
            clog_file = os.path.abspath(os.path.join(self.args.path, 'clog'))
            commit_msg_file = os.path.abspath(os.path.join(self.args.path, 'commit-message'))
            with open(commit_msg_file, 'w') as commit_msg:
                commit_msg.write(self.args.message)
                commit_msg.write('\n\n')
                with open(clog_file, 'r') as clog:
                    commit_msg.write(clog.read())
            self.args.file = commit_msg_file
            os.remove(clog_file)
            # This assignment is a magic because commit message is in the file
            # commit-message already.
            self.args.message = None

        try:
            self.cmd.commit(self.args.message, self.args.file, self.args.files, self.args.signoff)
            if self.args.tag:
                self.cmd.dump_spec_only = True
                self.cmd.version_bump = True
                self.cmd.add_tag()
        except Exception:
            if self.args.tag:
                self.log.error('Could not commit, will not tag!')
            if self.args.push:
                self.log.error('Could not commit, will not push!')
            raise
        finally:
            if self.args.with_changelog and os.path.isfile(self.args.file):
                os.remove(self.args.file)
                del self.args.file

        if self.args.push:
            self.push()

    def copr_build(self):
        srpm_path = self.srpm()

        if self.args.project:
            project = self.args.project
        else:
            project = self.cmd.ns_module_name

        self.cmd.copr_build(project,
                            srpm_path,
                            not self.args.wait,
                            self.args.copr_config)

    def is_packed(self):
        sys.stderr.write('This function is deprecated and will be removed in a future release.\n')
        self.cmd.dump_spec_only = True
        ts = rpm.ts()
        try:
            rpm.addMacro("_sourcedir", self.cmd.path)
            rpm_spec = ts.parseSpec(self.cmd.spec)
        except ValueError as e:
            raise RpmSpecParseException(str(e))

        if self.cmd.is_unpacked(self.cmd.path, rpm_spec.sources):
            self.log.info('No')
        else:
            self.log.info('Yes')

    def register_make_source(self):
        make_source_parser = self.subparsers.add_parser(
            name='make-source', help='Deprecated. Create Source0 from the '
            'content of the current working directory '
            'after downloading any external sources. '
            'The content must be of unpacked type.',
            description='Deprecated. Puts content of the current '
            'working directory into a gzip-compressed archive named '
            'according to Source0 filename as specified in the .spec file. '
            'The content must be of unpacked type, otherwise no action is taken. '
            'Unpacked content is such that it contains a .spec file '
            'that references no present source or patch '
            '(typically it contains only Source0 being '
            'generated automatically) and there is at least '
            'one file not in the list of ignored content (README, '
            'README.md, sources, LICENSE, tito.props, hidden files, '
            '.spec file). Note that by invoking this command '
            'with --outdir ./, the directory content becomes '
            '"packed".')
        make_source_parser.set_defaults(command=self.make_source)

    def register_spec(self):
        """Register the spec target"""
        spec_parser = self.subparsers.add_parser(
            name='spec', help='Generate spec file from a spec template',
            description='Preprocess a given spec template '
            'and put the resulting spec file into outdir. '
            'The spec template is simply copied into the outdir if '
            'spec preprocessing is turned off or if the input spec template '
            'does not contain any preprocessor macros.')
        spec_parser.add_argument(
            '--sources', action='store_true', default=False,
            help='Enable source generation by the input '
            'spec template. By default disabled.')
        spec_parser.set_defaults(command=self.spec)

    def register_is_packed(self):
        """Determine whether the package content is packed or not"""
        is_packed_parser = self.subparsers.add_parser(
            name='is-packed', help='Deprecated. Tell user whether content is packed',
            description='Deprecated. Determine whether the package content '
            'in the working directory is packed or unpacked '
            'and print that information to the screen.')
        is_packed_parser.set_defaults(command=self.is_packed)

    def register_lint(self):
        """Register the lint target"""
        lint_parser = self.subparsers.add_parser(
            'lint', help='Run rpmlint against local spec and build output if'
                         ' present.',
            description='Rpmlint can be configured using the --rpmlintconf/-r'
                         ' option or by setting a .rpmlint file in the'
                         ' working directory.')
        lint_parser.add_argument(
            '--info', '-i', default=False, action='store_true',
            help='Display explanations for reported messages')
        lint_parser.add_argument(
            '--rpmlintconf', '-r', default=None,
            help='Use a specific configuration file for rpmlint')
        lint_parser.add_argument(
            '--outdir', default=None,
            help='Directory with rpm items to be linted or just'
            ' a single rpm item like a .spec file. If not specified,'
            ' a directory from base output path will be selected.')
        lint_parser.set_defaults(command=self.lint)

    def register_srpm(self):
        """Register the srpm target"""
        srpm_parser = self.subparsers.add_parser(
            'srpm', help='Create a source rpm',
            parents=[self.rpm_parser_common],
            description='Create a source rpm')
        srpm_parser.set_defaults(command=self.srpm)

    def register_local(self):
        """Register the local target"""
        local_parser = self.subparsers.add_parser(
            'local', parents=[self.rpm_parser_common],
            help='Local test rpmbuild binary',
            description='Locally test run of rpmbuild producing binary RPMs.'
                        ' The rpmbuild output will be logged into a file named'
                        ' .build-%{version}-%{release}.log')
        local_parser.set_defaults(command=self.local)

    def register_verify_files(self):
        """Register the verify-files target"""
        verify_files_parser = self.subparsers.add_parser(
            'verify-files', help='Locally verify %%files section',
            description="Locally run 'rpmbuild -bl' to verify the spec file's"
                        " %files sections. This requires a successful run of "
                        "'{0} install' in advance.".format(self.name))
        verify_files_parser.add_argument(
            '--outdir', default=None,
            help='Output directory with build results.')
        verify_files_parser.set_defaults(command=self.verify_files)

    def register_copr_build(self):
        """Register the copr-build target"""
        description="""
Build package in Copr Build System.

You need to have copr-cli installed and also set up Copr api key.
For more information about the api key, see API KEY section in copr-cli(1)
man page.
"""
        copr_parser = self.subparsers.add_parser(
            'copr-build', help='Build package in Copr Build System',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=description)
        copr_parser.add_argument(
            '--config', required=False,
            metavar='CONFIG', dest='copr_config',
            help="Path to an alternative Copr configuration file")
        copr_parser.add_argument(
            '-w', '--wait',  action='store_true', default=False,
            help="Wait for build to finish.")
        copr_parser.add_argument(
            'project', nargs='?',
            help='Name of the project in format OWNER/PROJECT. If not'
            ' specified, project will be set to module name.')
        copr_parser.set_defaults(command=self.copr_build)

        build_parser = self.subparsers.add_parser(
            'build', parents=[copr_parser], conflict_handler='resolve',
            help='Alias for copr-build', description=description)
        build_parser.set_defaults(command=self.copr_build)

    def register_tag(self):
        """Register the tag target"""
        tag_parser = self.subparsers.add_parser(
            'tag', help='Management of git tags',
            description='This command uses git to create, list, or delete '
                        'tags.')
        tag_parser.add_argument(
            '-l', '--list', default=False, action='store_true',
            help='List all tags with a given pattern, or all if not pattern '
                 'is given')
        tag_parser.add_argument(
            '-d', '--delete', default=False, action='store_true',
            help='Delete a tag')
        tag_parser.add_argument(
            '-s', '--skip-edit', default=False, action='store_true',
            help='Skips changelog editing when creating a tag.')
        tag_parser.add_argument(
            'tag', nargs='?', default=None, help='Name of the tag')
        tag_parser.set_defaults(command=self.tag)

    def register_commit(self):
        """Register the commit target and ci alias"""
        commit_parser = self.subparsers.add_parser(
            'commit', help='Commit changes',
            description='This invokes a git commit. All tracked files with '
                        'changes will be committed unless a specific file '
                        'list is provided. $EDITOR will be used to generate a'
                        ' changelog message unless one is given to the '
                        'command. A tag or push can be done at the same time.')
        commit_parser.add_argument(
            '-m', '--message', default=None,
            help='Use the given <msg> as the commit message summary')
        commit_parser.add_argument(
            '--with-changelog',
            action='store_true',
            default=None,
            help='Get the last changelog from SPEC as commit message content. '
                 'This option must be used with -m together.')
        commit_parser.add_argument(
            '-t', '--tag', default=False, action='store_true',
            help='Create a tag for this commit')
        commit_parser.add_argument(
            '-F', '--file', default=None,
            help='Take the commit message from the given file')
        # allow one to commit /and/ push at the same time.
        commit_parser.add_argument(
            '-p', '--push', default=False, action='store_true',
            help='Commit and push as one action')
        # Allow a list of files to be committed instead of everything
        commit_parser.add_argument(
            'files', nargs='*', default=[],
            help='Optional list of specific files to commit')
        commit_parser.add_argument(
            '-s', '--signoff', default=False, action='store_true',
            help='Include a signed-off-by')
        commit_parser.set_defaults(command=self.commit)

        # Add a ci alias
        ci_parser = self.subparsers.add_parser(
            'ci', parents=[commit_parser], conflict_handler='resolve',
            help='Alias for commit')
        ci_parser.set_defaults(command=self.commit)

    def register_push(self):
        """Register the push target"""
        push_parser = self.subparsers.add_parser(
            'push', help='Push changes and annotated tags to remote repository.')
        push_parser.add_argument('--force', '-f', help='Force push', action='store_true')
        push_parser.set_defaults(command=self.push)

    def register_rpm_common(self):
        """Create a common parser for rpm commands"""
        self.rpm_parser_common = argparse.ArgumentParser(
            'rpm_common', add_help=False)
        self.rpm_parser_common.add_argument(
            '--target', default=None,
            help='Prep for a specific arch or arch-vendor-os.')
        self.rpm_parser_common.add_argument(
            '--with', metavar='OPTION', action='append', default=[],
            help='Enable %%bcond spec OPTION. Can be specified multiple times.')
        self.rpm_parser_common.add_argument(
            '--without', metavar='OPTION', action='append', default=[],
            help='Disable %%bcond spec OPTION. Can be specified multiple times.')

    def parse_cmdline(self, manpage=False):
        """Parse the commandline, optionally make a manpage

        This also sets up self.user
        """
        if manpage:
            # Generate the man page
            man_name = self.name
            if man_name.endswith('.py'):
                man_name = man_name[:-3]
            man_page = __import__('%s' % man_name)
            man_page.generate(self.parser, self.subparsers)
            sys.exit(0)
            # no return possible

        # Parse the args
        self.args = self.parser.parse_args()
        #command=self.parser.print_help

        if self.args.user:
            self.user = self.args.user
        else:
            self.user = getpass.getuser()
